/* jshint evil: true */
/**
 * Downloads JS files and runs it
 * @param {String} version - Version string of your game
 * @param {Function} onReady - Called when patcher has ended
 */
(function () {
    /*
     * Cocoon's atob function is unreliable
     * using atob method from https://gist.github.com/sevir/3946819
     */
    (function (window) {
        var _PADCHAR = "=",
            _ALPHA = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

        function _getbyte64(s, i) {
            var idx = _ALPHA.indexOf(s.charAt(i));

            if (idx === -1) {
                throw "Cannot decode base64";
            }

            return idx;
        }

        function _decode(s) {
            var pads = 0,
                i,
                b10,
                imax = s.length,
                x = [];

            s = String(s);

            if (imax === 0) {
                return s;
            }

            if (imax % 4 !== 0) {
                throw "Cannot decode base64";
            }

            if (s.charAt(imax - 1) === _PADCHAR) {
                pads = 1;

                if (s.charAt(imax - 2) === _PADCHAR) {
                    pads = 2;
                }

                // either way, we want to ignore this last block
                imax -= 4;
            }

            for (i = 0; i < imax; i += 4) {
                b10 = (_getbyte64(s, i) << 18) | (_getbyte64(s, i + 1) << 12) | (_getbyte64(s, i + 2) << 6) | _getbyte64(s, i + 3);
                x.push(String.fromCharCode(b10 >> 16, (b10 >> 8) & 0xff, b10 & 0xff));
            }

            switch (pads) {
            case 1:
                b10 = (_getbyte64(s, i) << 18) | (_getbyte64(s, i + 1) << 12) | (_getbyte64(s, i + 2) << 6);
                x.push(String.fromCharCode(b10 >> 16, (b10 >> 8) & 0xff));
                break;

            case 2:
                b10 = (_getbyte64(s, i) << 18) | (_getbyte64(s, i + 1) << 12);
                x.push(String.fromCharCode(b10 >> 16));
                break;
            }

            return x.join("");
        }

        function _getbyte(s, i) {
            var x = s.charCodeAt(i);

            if (x > 255) {
                throw "INVALID_CHARACTER_ERR: DOM Exception 5";
            }

            return x;
        }

        function _encode(s) {
            if (arguments.length !== 1) {
                throw "SyntaxError: exactly one argument required";
            }

            s = String(s);

            var i,
                b10,
                x = [],
                imax = s.length - s.length % 3;

            if (s.length === 0) {
                return s;
            }

            for (i = 0; i < imax; i += 3) {
                b10 = (_getbyte(s, i) << 16) | (_getbyte(s, i + 1) << 8) | _getbyte(s, i + 2);
                x.push(_ALPHA.charAt(b10 >> 18));
                x.push(_ALPHA.charAt((b10 >> 12) & 0x3F));
                x.push(_ALPHA.charAt((b10 >> 6) & 0x3f));
                x.push(_ALPHA.charAt(b10 & 0x3f));
            }

            switch (s.length - imax) {
            case 1:
                b10 = _getbyte(s, i) << 16;
                x.push(_ALPHA.charAt(b10 >> 18) + _ALPHA.charAt((b10 >> 12) & 0x3F) + _PADCHAR + _PADCHAR);
                break;

            case 2:
                b10 = (_getbyte(s, i) << 16) | (_getbyte(s, i + 1) << 8);
                x.push(_ALPHA.charAt(b10 >> 18) + _ALPHA.charAt((b10 >> 12) & 0x3F) + _ALPHA.charAt((b10 >> 6) & 0x3f) + _PADCHAR);
                break;
            }

            return x.join("");
        }

        window.encodeB64 = _encode;
        window.decodeB64 = _decode;

    })(window);

    var Xhr = function (settings) {
        /*settings = {
            get: String, // url to get
            post: String, // url to post
            params: {}, // additional parameters
            onSuccess: Function (res),
            onFail: Function (res),
        }*/
        var xhr = new window.XMLHttpRequest(),
            getPostBody = function () {
                // turns the params string into a url encoded form
                var param,
                    params = settings.params,
                    first = true,
                    url = '';
                for (param in params) {
                    if (!params.hasOwnProperty(param)) {
                        continue;
                    }

                    if (!first) {
                        url += '&';
                    }
                    url += param + '=' + params[param];
                    first = false;
                }
                return url;
            },
            addHeaders = function () {
                var key,
                    headers = settings.headers;
                for (key in headers) {
                    if (!headers.hasOwnProperty(key)) {
                        continue;
                    }
                    xhr.setRequestHeader(key, headers[key]);
                }
            };

        if (settings.get) {
            xhr.open('GET', settings.get, true);
        } else if (settings.post) {
            xhr.open('POST', settings.post, true);
            if (settings.json) {
                xhr.setRequestHeader('Content-Type', 'text/plain');
                xhr.setRequestHeader('X-Requested-With', 'XMLHttpRequest');
            } else {
                xhr.setRequestHeader('Content-type', 'application/x-www-form-urlencoded');
            }
        }

        if (settings.headers) {
            addHeaders();
        }

        xhr.onreadystatechange = function () {
            if (xhr.readyState === 4) {
                if (xhr.status >= 200 && xhr.status < 300) {
                    if (!settings.onSuccess) {
                        return;
                    }
                    try {
                        settings.onSuccess(JSON.parse(xhr.responseText));
                    } catch (e) {
                        settings.onSuccess({
                            response: xhr.responseText
                        });
                    }
                } else {
                    if (!settings.onFail) {
                        return;
                    }
                    try {
                        settings.onFail(JSON.parse(xhr.responseText));
                    } catch (e) {
                        settings.onFail({
                            response: xhr.responseText
                        });
                    }
                }
            }
        };

        //console.log('success', xhr.responseText);
        if (settings.json) {
            xhr.send(JSON.stringify(settings.params));
        } else if (settings.post) {
            xhr.send(getPostBody());
        } else {
            xhr.send();
        }

        return xhr;
    };
    window.patcher = function (baseUrl, version, onReady) {
        var timeout = 5000, // should complete in 5 seconds
            hasTimedout = false,
            hasFinished = false,
            currentPatch,
            // step 1
            downloadSettings = function () {
                if (hasTimedout) {
                    // too late!
                    return;
                }
                new Xhr({
                    get: baseUrl + '/settings.json',
                    onSuccess: function (res) {
                        var patchVersion = res[version] || 0;
                        console.log('Detected ' + patchVersion);

                        if (patchVersion === -1) {
                            // wipe patch!
                            localStorage.setItem('patch_' + version, null);
                            onReady('Wiped clean.');
                        } else if (patchVersion > currentPatch.version) {
                            // download new patch
                            console.log('Download new ' + patchVersion);
                            downloadPatch(patchVersion);
                        } else {
                            // equal patch, apply current
                            console.log('Apply current ' + currentPatch.version);
                            applyPatch();
                        }
                    },
                    onFail: applyPatch
                });
            },
            // step 2
            downloadPatch = function (patchVersion) {
                if (hasTimedout) {
                    // too late!
                    return;
                }
                var fileName = version + '-' + patchVersion + '.json';
                new Xhr({
                    get: baseUrl + '/' + fileName,
                    onSuccess: function (res) {
                        // store patch
                        currentPatch.version = patchVersion;
                        currentPatch.data = res.data;
                        currentPatch.method = res.method || 'default';

                        localStorage.setItem('patch_' + version, JSON.stringify(currentPatch));
                        applyPatch();
                    },
                    onFail: applyPatch
                });
            },
            // step 3
            applyPatch = function () {
                var code;

                if (hasFinished) {
                    // already applied the patch (possibly due to time out)
                    return;
                }
                hasFinished = true;

                // there is no patch
                if (!currentPatch.data) {
                    onReady('Nothing to do.');
                    return;
                }

                // !!
                try {
                    // decompress code
                    if (currentPatch.method === 'atob') {
                        // using atob could be unreliable because of Cocoon
                        code = window.atob(currentPatch.data);
                    } else {
                        // default method is the javascript method
                        code = window.decodeB64(currentPatch.data);
                    }
                    eval(code);
                    window.PATCHVERSION = currentPatch.version;
                    onReady('Success');
                } catch (e) {
                    onReady('Failed with', e);
                }
            };

        window.setTimeout(function () {
            if (hasFinished) {
                return;
            }
            hasTimedout = true;
            console.log('WARNING: Download timed out.');
            applyPatch();
        }, timeout);

        // load currentpatch
        try {
            // try to parse the current patch
            currentPatch = JSON.parse(localStorage.getItem('patch_' + version)) || {
                version: 0
            };
        } catch (e) {
            // something went horribly wrong...
            currentPatch = {
                version: 0
            };
        }

        // start by trying to download settings
        downloadSettings();
    };

})();